from typing import Union, List, Optional
from decimal import Decimal
from datetime import date, time, datetime
from enum import Enum
from . import validators, formatter, exceptions

class CNABFieldType(Enum):
    Decimal = 'decimal'
    Int = 'int'
    Alfa = 'alfa'
    Alfa2 = 'alfa2'
    Date = 'date'
    Time = 'time'
    DateReverse = 'dateReverse'

class CNABFieldValueError(Exception):
    pass

CNABFieldValueType = Union[str, date, time, datetime, int, float, Decimal]

class CNABField:
    name: str = ''
    length: int = 0
    validation: CNABFieldType
    default: CNABFieldValueType
    required: bool
    validators: List = []
    precision: int = 0
    formatter = None

    def __init__(
        self,
        length: int,
        validation: CNABFieldType,
        default: CNABFieldValueType,
        required: Optional[bool] = False,
        precision: Optional[int] = 2,
    ):
        self.length = length
        self.validation = validation
        self.default = default
        self.required = required
        self.precision = precision
        self.formatter = None

        self.validators = []
        if self.required:
            self.validators += [validators.validate_required]
        
        if self.validation == CNABFieldType.Int:
            self.validators += [validators.validate_integer]
            self.formatter = formatter.format_integer

        if self.validation == CNABFieldType.Decimal:
            self.validators += [validators.validate_decimal]
            self.formatter = formatter.format_decimal

        if self.validation == CNABFieldType.Alfa:
            self.formatter = formatter.format_alfa
            
        if self.validation == CNABFieldType.Alfa2:
            self.formatter = formatter.format_alfa2

        if self.validation == CNABFieldType.Date:
            self.validators += [validators.validate_date]
            self.formatter = formatter.format_date

        if self.validation == CNABFieldType.Time:
            self.validators += [validators.validate_date]
            self.formatter = formatter.format_time

        if not self.formatter:
            raise exceptions.CNABFieldTypeNotSupportedError(self.validation)
        
    def validate_value(self, value: CNABFieldValueType):
        for validator in self.validators:
            value = validator(value, self)
        return value

    def format_value(self, value: CNABFieldValueType):
        value = self.validate_value(value)
        return self.formatter(value, self)



from typing import List
from abc import ABC
from .cnab_240.registro0 import CNAB240Registro0

class Remessa(ABC):
    header: CNAB240Registro0
    registro0_class = None
    registro9_class = None

    def __init__(self, **kwargs):
        self.header = self.registro0_class(**kwargs)

    def inserir_lote(self, **kwargs):
        return self.header.inserir_lote(**kwargs)

    def get_text(self) -> List[str]:
        footer = self.registro9_class(self.header, None, **{ '1': 1 })
        return self.header.get_text() + footer.get_text()



from cnab.base.cnab_240 import CNAB240Registro1
from cnab.core.field import CNABField, CNABFieldType
from .registro5 import Santander240Registro5

class Santander240Registro1(CNAB240Registro1):
    registro5_class = Santander240Registro5
    _meta = {
        "codigo_banco": CNABField(
            length=3, default="033", validation=CNABFieldType.Int, required=True
        ),
        "codigo_lote": CNABField(
            length=4, default=1, validation=CNABFieldType.Int, required=True
        ),
        "tipo_registro": CNABField(
            length=1, default=1, validation=CNABFieldType.Int, required=True
        ),
        "operacao": CNABField(
            length=1, default="R", validation=CNABFieldType.Alfa, required=True
        ),
        "tipo_servico": CNABField(
            length=2, default="01", validation=CNABFieldType.Int, required=True
        ),
        "filler1": CNABField(
            length=2, default=" ", validation=CNABFieldType.Alfa, required=True
        ),
        "versao_layout": CNABField(
            length=3, default="030", validation=CNABFieldType.Int, required=True
        ),
        "filler2": CNABField(
            length=1, default=" ", validation=CNABFieldType.Alfa, required=True
        ),
        "tipo_inscricao": CNABField(
            length=1, default="", validation=CNABFieldType.Int, required=True
        ),
        "numero_inscricao": CNABField(
            length=15, default="", validation=CNABFieldType.Int, required=True
        ),
        "filler3": CNABField(
            length=20, default=" ", validation=CNABFieldType.Alfa, required=True
        ),
        "agencia": CNABField(
            length=4, default="", validation=CNABFieldType.Int, required=True
        ),
        "filler12": CNABField(
            length=4, default="0", validation=CNABFieldType.Int, required=True
        ),
        "codigo_beneficiario": CNABField(
            length=6, default="0", validation=CNABFieldType.Int, required=True
        ),
        "codigo_beneficiario_dv": CNABField(
            length=1, default="0", validation=CNABFieldType.Int, required=True
        ),
        "filler4": CNABField(
            length=5, default=" ", validation=CNABFieldType.Alfa, required=True
        ),
        "nome_empresa": CNABField(
            length=30, default="", validation=CNABFieldType.Alfa, required=True
        ),
        "mensagem1": CNABField(
            # mensagems 1 e 2 : somente use para mensagens que serao impressas de forma identica em todos os boletos do lote
            length=40,
            default=" ",
            validation=CNABFieldType.Alfa,
            required=True,
        ),
        "mensagem2": CNABField(
            # mensagems 1 e 2 : somente use para mensagens que serao impressas de forma identica em todos os boletos do lote
            length=40,
            default=" ",
            validation=CNABFieldType.Alfa,
            required=True,
        ),
        "numero_remessa": CNABField(
            length=8, default="", validation=CNABFieldType.Int, required=True
        ),
        "data_gravacao": CNABField(
            length=8,
            default="",  # nao informar a data na instanciação - gerada dinamicamente
            validation=CNABFieldType.Date,
            required=True,
        ),
        "filler5": CNABField(
            length=41, default=" ", validation=CNABFieldType.Alfa, required=True
        ),
    }

    def inserir_detalhe(self, **kwargs):
        from .registro3P import Santander240Registro3P
        return Santander240Registro3P(self.header, self, self, **kwargs)


import re
from unidecode import unidecode

def remove_accents(text: str) -> str:
    regexes = [
        r'\xc3[\x80-\x85]',
        r'\xc3\x87',
        r'\xc3[\x88-\x8b]',
        r'\xc3[\x8c-\x8f]',
        r'\xc3([\x92-\x96]|\x98)',
        r'\xc3[\x99-\x9c]',
        r'\xc3[\xa0-\xa5]',
        r'\xc3\xa7',
        r'\xc3[\xa8-\xab]',
        r'\xc3[\xac-\xaf]',
        r'\xc3([\xb2-\xb6]|\xb8)',
        r'\xc3[\xb9-\xbc]'
    ]
    replacers = ['A', 'C', 'E', 'I', 'O', 'U', 'a', 'c', 'e', 'i', 'o', 'u']
    for regex, repl in zip(regexes, replacers):
        text = re.sub(regex, repl, text).lower()
    return text

def prepare_text(text: str) -> str:
    return unidecode(text.strip()).upper()
    return remove_accents(text.strip()).upper()

def format_alfa(value, field):
    value = prepare_text(str(value))
    if len(value) > field.length:
        value = value[0:field.length]
    return value.ljust(field.length, ' ')



from cnab.base.cnab_240 import CNAB240Registro3
from cnab.core.field import CNABField, CNABFieldType

class Santander240Registro3R(CNAB240Registro3):
    _meta = {
        'codigo_banco': CNABField(
            length=3,
            default='033',
            validation=CNABFieldType.Int,
            required=True
        ),
        'codigo_lote': CNABField(
            length=4,
            default=1,
            validation=CNABFieldType.Int,
            required=True
        ),
        'tipo_registro': CNABField(
            length=1,
            default='3',
            validation=CNABFieldType.Int,
            required=True
        ),
        'numero_registro': CNABField(
            length=5,
            default='0',
            validation=CNABFieldType.Int,
            required=True
        ),
        'seguimento': CNABField(
            length=1,
            default='R',
            validation=CNABFieldType.Alfa,
            required=True
        ),
        'filler1': CNABField(
            length=1,
            default=' ',
            validation=CNABFieldType.Alfa,
            required=True
        ),
        'codigo_movimento': CNABField(
            length=2,
            default='01', # entrada de titulo
            validation=CNABFieldType.Int,
            required=True
        ),
        # - ------------------ ate aqui é igual para todo registro tipo 3
        'codigo_desconto2': CNABField(
            length=1,
            default='0',
            validation=CNABFieldType.Int,
            required=True
        ),
        'data_desconto2': CNABField(
            length=8,
            default='0',
            validation=CNABFieldType.Date,
            required=True
        ),
        'vlr_desconto2': CNABField(
            length=13,
            default='0',
            validation=CNABFieldType.Decimal,
            precision=2,
            required=True
        ),
        'filler14': CNABField(
            length=24,
            default=' ',
            validation=CNABFieldType.Alfa,
            required=True
        ),
        'codigo_multa': CNABField(
            length=1,
            default='1',
            validation=CNABFieldType.Int,
            required=True
        ),
        'data_multa': CNABField(
            length=8,
            default='0',
            validation=CNABFieldType.Date,
            required=True
        ),
        'vlr_multa': CNABField(
            length=13,
            default='0',
            validation=CNABFieldType.Decimal,
            precision=2,
            required=True
        ),
        'filler15': CNABField(
            length=10,
            default=' ',
            validation=CNABFieldType.Alfa,
            required=True
        ),
        'mensagem3': CNABField(
            length=40,
            default=' ',
            validation=CNABFieldType.Alfa,
            required=True
        ),
        'mensagem4': CNABField(
            length=40,
            default=' ',
            validation=CNABFieldType.Alfa,
            required=True
        ),
        'filler16': CNABField(
            length=61,
            default=' ',
            validation=CNABFieldType.Alfa,
            required=True
        ),
    }



from typing import Optional
from cnab.base.registro import Registro
from cnab.base.registro_remessa import RegistroRemessa
from .registro1 import CNAB240Registro1

class CNAB240Registro5(RegistroRemessa):
    lote: CNAB240Registro1 = None

    def __init__(self,
        header: Optional["Registro"],
        parent: Optional["Registro"],
        lote: CNAB240Registro1,
        **kwargs: dict,
    ):
        self.lote = lote
        super().__init__(header, parent, **kwargs)

    def get_qtd_registros(self):
        return self.lote.counter + 2



from .registro import Registro

class RegistroRemessa(Registro):
    def get_text(self) -> str:
        retorno = ''
        for key, field in self._meta.items():
            retorno += self.get_value(key, field.default)
            
        result = [retorno]

        for child in self._children:
            result += child.get_text()

        return result



from typing import Optional
from datetime import datetime
from cnab.base.registro_remessa import RegistroRemessa
from cnab.base.registro import Registro
from cnab.core.enums import TipoServico, TipoInscricao
from cnab.core.exceptions import CNABInvalidTypeError

class CNAB240Registro1(RegistroRemessa):
    def __init__(self, header: Optional["Registro"], parent: Optional["Registro"], **kwargs: dict):
        self.counter = 0
        super().__init__(header, parent, **kwargs)

    def set_codigo_lote(self):
        self._data['codigo_lote'] = self.header.counter

    def set_tipo_servico(self, value: TipoServico):
        if not isinstance(value, TipoServico):
            raise CNABInvalidTypeError(TipoServico)

        self._data['tipo_servico'] = value.value

    def set_tipo_inscricao(self, value: TipoInscricao):
        if not isinstance(value, TipoInscricao):
            raise CNABInvalidTypeError(TipoInscricao)
        
        self._data['tipo_inscricao'] = value.value

    def get_tipo_servico(self):
        return self.get_data_or_parent('tipo_servico')
    
    def get_tipo_inscricao(self):
        return self.get_data_or_parent('tipo_inscricao')
    
    def get_numero_inscricao(self):
        return self.get_data_or_parent('numero_inscricao')
    
    def get_agencia(self):
        return self.get_data_or_parent('agencia')
    
    def get_agencia_dv(self):
        return self.get_data_or_parent('agencia_dv')
    
    def get_conta(self):
        return self.get_data_or_parent('conta')
    
    def get_conta_dv(self):
        return self.get_data_or_parent('conta_dv')
    
    def get_convenio(self):
        return self.get_data_or_parent('convenio')
    
    def get_nome_empresa(self):
        return self.get_data_or_parent('nome_empresa')
    
    def get_numero_remessa(self):
        return self.get_data_or_parent('numero_sequencial_arquivo')
    
    def get_data_gravacao(self):
        return datetime.now()

    def get_codigo_beneficiario(self):
        return self.get_data_or_parent('codigo_beneficiario')
    
    def get_codigo_beneficiario_dv(self):
        return self.get_data_or_parent('codigo_beneficiario_dv')
    
    def get_text(self) -> str:
        dataReg5 = {}
        dataReg5['qtd_titulos_simples'] = 0
        dataReg5['qtd_titulos_caucionada'] = 0
        dataReg5['qtd_titulos_descontada'] = 0
        dataReg5['vrl_titulos_simples'] = 0.00
        dataReg5['vlr_titulos_caucionada'] = 0.00
        dataReg5['vlr_titulos_descontada'] = 0.00

        retorno = ''
        for key, field in self._meta.items():
            retorno += self.get_value(key, field.default)

        result = [retorno]

        if self._children:
            for child in self._children:
                if child.get_codigo_carteira() == 1:
                    dataReg5['qtd_titulos_simples'] += 1
                    dataReg5['vrl_titulos_simples'] += float(child.get_unformated('valor'))
                if child.get_codigo_carteira() == 3:
                    dataReg5['qtd_titulos_caucionada'] += 1
                    dataReg5['vlr_titulos_caucionada'] += float(child.get_unformated('valor'))
                if child.get_codigo_carteira() == 4:
                    dataReg5['qtd_titulos_descontada'] += 1
                    dataReg5['vlr_titulos_descontada'] += float(child.get_unformated('valor'))
                result += child.get_text()

            if not hasattr(self, 'registro5_class'):
                raise Exception("TODO: change the text of this exception")
            if not self.registro5_class:
                raise Exception("TODO: change the text of this exception")
                
            reg5 = self.registro5_class(None, None, self, **dataReg5)
            result += reg5.get_text()

        return result
    
    def append(self, child: Registro):
        super().append(child)

        self.counter += 1



from typing import Optional, Union, TYPE_CHECKING
from cnab.base.remessa import Remessa
from cnab.utils.dict_utils import set_if_has_value
from .registro0 import BancoBrasil240Registro0
from .registro9 import BancoBrasil240Registro9

if TYPE_CHECKING:
    from .registro1 import BancoBrasil240Registro1

class CNAB240BancoBrasil(Remessa):
    header: BancoBrasil240Registro0
    registro0_class = BancoBrasil240Registro0
    registro9_class = BancoBrasil240Registro9

    def __init__(
        self,
        *,
        codigo_lote: Optional[Union[str, int]] = None,
        tipo_registro: Optional[Union[str, int]] = None,
        tipo_inscricao: Optional[Union[str, int]] = None,
        numero_inscricao: Optional[Union[str, int]] = None,
        convenio: Optional[Union[str, int]] = None,
        cobranca_cedente: Optional[Union[str, int]] = None,
        carteira_cobranca: Optional[Union[str, int]] = None,
        variacao_carteira_cobranca: Optional[Union[str, int]] = None,
        agencia: Optional[Union[str, int]] = None,
        agencia_dv: Optional[Union[str, int]] = None,
        conta: Optional[Union[str, int]] = None,
        conta_dv: Optional[Union[str, int]] = None,
        nome_empresa: Optional[str] = None,
        codigo_remessa: Optional[Union[str, int]] = None,
        numero_sequencial_arquivo: Optional[Union[str, int]] = None,
        **kwargs,
    ):
        set_if_has_value(kwargs, 'codigo_lote', codigo_lote)
        set_if_has_value(kwargs, 'tipo_registro', tipo_registro)
        set_if_has_value(kwargs, 'tipo_inscricao', tipo_inscricao)
        set_if_has_value(kwargs, 'numero_inscricao', numero_inscricao)
        set_if_has_value(kwargs, 'convenio', convenio)
        set_if_has_value(kwargs, 'cobranca_cedente', cobranca_cedente)
        set_if_has_value(kwargs, 'carteira_cobranca', carteira_cobranca)
        set_if_has_value(kwargs, 'variacao_carteira_cobranca', variacao_carteira_cobranca)
        set_if_has_value(kwargs, 'agencia', agencia)
        set_if_has_value(kwargs, 'agencia_dv', agencia_dv)
        set_if_has_value(kwargs, 'conta', conta)
        set_if_has_value(kwargs, 'conta_dv', conta_dv)
        set_if_has_value(kwargs, 'nome_empresa', nome_empresa)
        set_if_has_value(kwargs, 'codigo_remessa', codigo_remessa)
        set_if_has_value(kwargs, 'numero_sequencial_arquivo', numero_sequencial_arquivo)
        super().__init__(**kwargs)

    def inserir_lote(
        self,
        *,
        tipo_servico: Optional[Union[str, int]] = None,
        variacao: Optional[Union[str, int]] = None,
        **kwargs,
    ) -> "BancoBrasil240Registro1":
        set_if_has_value(kwargs, 'tipo_servico', tipo_servico)
        set_if_has_value(kwargs, 'variacao', variacao)
        return super().inserir_lote(**kwargs)



from cnab.core.exceptions import CNABFieldNotIntegerError

def validate_integer(value, field):
    try:
        value = int(str(value))
    except Exception:
        raise CNABFieldNotIntegerError(field.name)
    return value



from abc import abstractmethod
from enum import Enum
from typing import Union
from cnab.core.exceptions import CNABInvalidEspecieTituloError
from .banks import BANKS

class EspecieTitulo(Enum):
    Cheque = 'CH'
    DuplicataMercantil = 'DM'
    MercantilParaIndicacao = 'DMI'
    DuplicataDeServico = 'DS'
    DuplicataDeServicoParaIndicacao = 'DSI'
    DuplicataRural = 'DR'
    LetraCambio = 'LC'
    NotaCreditoComercial = 'NCC'
    NotaCreditoExportacao = 'NCE'
    NotaCreditoExportacaoIndicacao = 'NCI'
    NotaCreditoRural = 'NCR'
    NotaPromissoria = 'NP'
    NotaPromissoriaRural = 'NPR'
    TriplicataMercantil = 'TM'
    TriplicataServico = 'TS'
    NotaSeguro = 'NS'
    Recibo = 'RC'
    Fatura = 'FAT'
    NotaDebito = 'ND'
    Warrant = 'WRT'
    ApoliceSeguros = 'AP'
    MensalidadeEscolar = 'ME'
    ParcelaConsorcio = 'ME'
    Contrato = 'CT'
    Cosseguros = 'CS'
    DocumentoDivida = 'DV'
    EncargosCondominiais = 'EC'
    ContaPrestacaoServicoes = 'CPS'
    BoletoProposta = 'DBP'
    CartaoCredito = 'CC'
    Outros = 'DIV'

    @abstractmethod
    def get_real_value(bank: Union[str, int], value: Union[str, "EspecieTitulo"]):
        default_value = value
        if isinstance(value, EspecieTitulo):
            value = value.value
        
        if not BANKS.get(str(bank)):
            raise CNABInvalidEspecieTituloError(default_value, bank)
        codes = BANKS.get(str(bank)).values()
        if value in codes:
            return value
        code = BANKS.get(str(bank)).get(value)
        if not code:
            raise CNABInvalidEspecieTituloError(default_value, bank)
        return code
    
__all__ = ['EspecieTitulo']



from cnab.base.cnab_240 import CNAB240Registro5
from cnab.core.field import CNABField, CNABFieldType


class BancoBrasil240Registro5(CNAB240Registro5):
    _meta = {
        "codigo_banco": CNABField(  # 01.5
            length=3, default="001", validation=CNABFieldType.Int, required=True
        ),
        "codigo_lote": CNABField(  # 02.5
            length=4, default=1, validation=CNABFieldType.Int, required=True
        ),
        "tipo_registro": CNABField(  # 03.5
            length=1, default="5", validation=CNABFieldType.Int, required=True
        ),
        "filler1": CNABField(  # 04.5
            length=9, default=" ", validation=CNABFieldType.Alfa, required=True
        ),
        "qtd_registros": CNABField(  # 05.5
            length=6, default="0", validation=CNABFieldType.Int, required=True
        ),
        "filler3": CNABField(  # 13.5
            length=217, default=" ", validation=CNABFieldType.Alfa, required=True
        ),
    }



from .field import (
    CNABFieldValueRequiredError,
    CNABFieldNotIntegerError,
    CNABFieldNotDecimalError,
    CNABFieldTypeNotSupportedError,
    IsNotCNABFieldError,
    IsNotValidDateTimeError,
)

from .types import CNABInvalidTypeError

from .especie import CNABInvalidEspecieTituloError

__all__ = [
    'CNABFieldValueRequiredError',
    'CNABFieldNotIntegerError',
    'CNABFieldNotDecimalError',
    'CNABFieldTypeNotSupportedError',
    'IsNotCNABFieldError',
    'IsNotValidDateTimeError',

    'CNABInvalidTypeError',

    'CNABInvalidEspecieTituloError',
]


from cnab.base.remessa import Remessa
from .registro0 import Santander240Registro0
from .registro9 import Santander240Registro9

class CNAB240Santander(Remessa):
    header: Santander240Registro0
    registro0_class = Santander240Registro0
    registro9_class = Santander240Registro9

    def inserir_lote(self, **kwargs):
        # TODO: implement kwargs named here
        return super().inserir_lote(**kwargs)



from cnab.base.registro_remessa import RegistroRemessa

class CNAB240Registro9(RegistroRemessa):
    def get_qtd_lotes(self):
        return self.header.counter
    
    def get_qtd_registros(self):
        last_lote = self.header._children[-1]
        return last_lote.counter + 4



from typing import Optional
from cnab.base.registro_remessa import RegistroRemessa
from cnab.base.registro import Registro
from cnab.core.especie import EspecieTitulo
from .registro1 import CNAB240Registro1

class CNAB240Registro3(RegistroRemessa):
    lote: CNAB240Registro1 = None

    def __init__(self,
        header: Optional["Registro"],
        parent: Optional["Registro"],
        lote: CNAB240Registro1,
        **kwargs: dict,
    ):
        self.lote = lote
        super().__init__(header, parent, **kwargs)
        self.init_numero_registro()

    def init_numero_registro(self):
        self._data['numero_registro'] = self.lote.counter
        #return self.lote.counter

    def get_agencia(self):
        return self.get_data_or_parent('agencia')
    
    def get_agencia_dv(self):
        return self.get_data_or_parent('agencia_dv')
    
    def get_conta(self):
        return self.get_data_or_parent('conta')

    def get_conta_dv(self):
        return self.get_data_or_parent('conta_dv')
    
    def get_codigo_beneficiario(self):
        return self.get_data_or_parent('codigo_beneficiario')

    def get_codigo_beneficiario_dv(self):
        return self.get_data_or_parent('codigo_beneficiario_dv')
    
    def get_cep_pagador(self):
        cep = self.get_data_or_parent('cep_pagador')
        
        if self._meta.get('cep_sufixo'):
            return cep.split('-')[0][0:5]
        
        return cep.replace('-', '')
    
    def get_cep_sufixo(self):
        cep = self.get_data_or_parent('cep_pagador')
        if "-" in cep:
            return cep.split('-')[-1]
        return cep[-3:]
    
    def get_seu_numero2(self):
        value = self._data.get('seu_numero2')
        if value:
            return value
        
        if hasattr(self, 'get_nosso_numero'):
            return self.get_nosso_numero()
        
        return self._data.get('nosso_numero')
    
    def append(self, child: Registro):
        super().append(child)
        self.lote.counter += 1

    def get_especie_titulo(self):
        field = self.get_field('codigo_banco')
        return EspecieTitulo.get_real_value(
            field.default,
            self._data.get('especie_titulo')
        )
    